import math
import numpy as np


def matrix(roll: float, pitch: float, yaw: float) -> np.array:
    salpha, calpha = math.sin(roll), math.cos(roll)
    sbeta, cbeta = math.sin(pitch), math.cos(pitch)
    sgamma, cgamma = math.sin(yaw), math.cos(yaw)
    a11 = cgamma * cbeta
    a12 = -sgamma * cbeta
    a13 = sbeta
    a21 = cgamma * sbeta * salpha + sgamma * calpha
    a22 = -sgamma * sbeta * salpha + cgamma * calpha
    a23 = -cbeta * salpha
    a31 = -cgamma * sbeta * calpha + sgamma * salpha
    a32 = sgamma * sbeta * calpha + cgamma * salpha
    a33 = cbeta * calpha
    return np.array([[a11, a12, a13], [a21, a22, a23], [a31, a32, a33]])



import numpy as np
from ptc.types import Point3f


def _calculate_coefficient_matrix(
    p1: Point3f, p2: Point3f, p3: Point3f, p4: Point3f
) -> np.array:
    x1, y1, z1 = p1
    x2, y2, z2 = p2
    x3, y3, z3 = p3
    x4, y4, z4 = p4
    return np.array(
        [
            [x1 - x2, y1 - y2, z1 - z2],
            [x2 - x3, y2 - y3, z2 - z3],
            [x3 - x4, y3 - y4, z3 - z4],
        ]
    )


def _calculate_ordinate_values(
    p1: Point3f, p2: Point3f, p3: Point3f, p4: Point3f
) -> np.array:
    x1, y1, z1 = p1
    x2, y2, z2 = p2
    x3, y3, z3 = p3
    x4, y4, z4 = p4
    return np.array(
        [
            ((x2 * x2 + y2 * y2 + z2 * z2) - (x1 * x1 + y1 * y1 + z1 * z1))
            / 2,
            ((x3 * x3 + y3 * y3 + z3 * z3) - (x2 * x2 + y2 * y2 + z2 * z2))
            / 2,
            ((x4 * x4 + y4 * y4 + z4 * z4) - (x3 * x3 + y3 * y3 + z3 * z3))
            / 2,
        ]
    )


def calculate_center(
    p1: Point3f, p2: Point3f, p3: Point3f, p4: Point3f
) -> Point3f:
    a = _calculate_coefficient_matrix(p1, p2, p3, p4)
    b = _calculate_ordinate_values(p1, p2, p3, p4)
    xs, ys, zs = np.linalg.solve(a, b)
    return (-xs, -ys, -zs)



import numpy as np
from hamcrest import *
from ptc import sphere


def test_sphere_center_close(valid_data_point):
    points = [p[0] for p in valid_data_point["input"]][:4]
    expected = valid_data_point["sphere_center"]
    got = sphere.calculate_center(*points)
    center_is_close = np.allclose(np.array(got), np.array(expected), atol=1e-3)
    assert_that(center_is_close, is_(True))



import numpy as np
import scipy.optimize
from typing import Callable

from ptc import sphere, rotation, types


def model(
    point_rotations: types.CalibrationInput, coeffs: np.array
) -> np.array:
    result = [
        np.array(pr[0]) + np.dot(rotation.matrix(*pr[1]), coeffs)
        for pr in point_rotations
    ]
    return np.array(result)


def residual_fn(
    optimization_target: np.array,
) -> Callable[[np.array, types.CalibrationInput], np.array]:
    def fn(
        coeffs: np.array, point_rotations: types.CalibrationInput
    ) -> np.array:
        model_result = model(point_rotations, coeffs)
        residuals = (optimization_target - model_result) ** 2
        return np.sum(residuals, axis=0)

    return fn


def calibrate(
    data_points: types.CalibrationInput,
    initial_guess: types.Point3f = (0.1, 0.1, 0.1),
) -> scipy.optimize.OptimizeResult:
    if len(data_points) < 4:
        raise ValueError("Not enough data points (less than 4)")

    points = [d[0] for d in data_points][:4]
    sphere_center = np.array(sphere.calculate_center(*points))
    initial_coeffs = np.array(initial_guess)
    optimization_target = np.array(
        [sphere_center for _ in range(len(data_points))]
    )
    optimization_result = scipy.optimize.least_squares(
        residual_fn(optimization_target),
        initial_coeffs,
        args=(data_points,),
        ftol=5e-9,
        xtol=5e-9,
        jac="3-point",
        loss="soft_l1",
    )
    return optimization_result



from typing import List
import pathlib

from setuptools import setup, find_packages

here = pathlib.Path(__file__).parent


def read_version() -> str:
    file_path = here / "version"
    with open(file_path) as version_file:
        return version_file.read().strip()


def read_requirements(path: str) -> List[str]:
    file_path = here / path
    with open(file_path) as requirements_file:
        return requirements_file.read().split("\n")


def development_status(version: str) -> str:
    if "a" in VERSION:
        dev_status = "Development Status :: 3 - Alpha"
    elif "dev" in VERSION:
        dev_status = "Development Status :: 4 - Beta"
    else:
        dev_status = "Development Status :: 5 - Production/Stable"
    return dev_status


def long_description(short_description: str) -> str:
    readme_path = here / "README.md"
    try:
        with open(readme_path,  encoding="utf-8") as readme:
            long_description = "\n" + readme.read()
            return long_description
    except FileNotFoundError:
        return short_description



NAME = "ptc"
DESCRIPTION = "Pulse Tool Calibration library"
URL = "https://rozum.com"
EMAIL = "dev@rozum.com"
AUTHOR = "Rozum Robotics"
VERSION = read_version()
DEVELOPMENT_STATUS = development_status(VERSION)
REQUIRED = read_requirements("requirements/production.txt")
LONG_DESCRIPTION = long_description(DESCRIPTION)


setup(
    name=NAME,
    version=VERSION,
    packages=find_packages(exclude=("test")),
    install_requires=REQUIRED,
    url=URL,
    license="Apache License 2.0",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        DEVELOPMENT_STATUS,
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    author=AUTHOR,
    author_email=EMAIL,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    zip_safe=False,
    
)



import numpy as np
from hamcrest import *
import ptc

def test_valid_calibration(valid_data_point):
    inp = valid_data_point["input"]
    expected = valid_data_point["tcp_offset"]
    got = ptc.calibrate(inp)
    solution_is_close = np.allclose(got.x, np.array(expected), atol=1e-3)
    assert_that(solution_is_close, is_(True))


def test_valid_output_length(valid_data_point):
    inp = valid_data_point["input"]
    got = ptc.calibrate(inp)
    assert_that(got.x, has_length(3))


